"""Functions to work with files and data formats."""

import os
import shutil
import logging

import simplejson as json
import yaml

from . import exceptions

log = logging.getLogger(__name__)


def exists(path):
    """Determine if a path exists."""
    return os.path.exists(path)


def touch(path):
    """Ensure a file path exists."""
    if not os.path.exists(path):
        dirpath = os.path.dirname(path)
        if dirpath and not os.path.isdir(dirpath):
            log.trace("Creating directory '{}'...".format(dirpath))
            os.makedirs(dirpath)
        log.trace("Creating empty '{}'...".format(path))
        write("", path)


def read(path, encoding='utf-8'):
    """Read text from a file.

    :param path: file path to read from
    :param encoding: input file encoding

    :return: string contents of file

    """
    log.trace("Reading text from '{}'...".format(path))

    with open(path, 'r', encoding=encoding) as stream:
        text = stream.read()

    return text


def write(text, path, encoding='utf-8'):
    """Write text to a file.

    :param text: string
    :param path: file path to write text
    :param encoding: output file encoding

    :return: path of file

    """
    if text:
        log.trace("Writing text to '{}'...".format(path))

    with open(path, 'wb') as stream:
        data = text.encode(encoding)
        stream.write(data)

    return path


def stamp(path):
    """Get the modification timestamp from a file."""
    return os.path.getmtime(path)


def delete(path):
    """Delete a file or directory."""
    if os.path.isdir(path):
        try:
            log.trace("Deleting '{}'...".format(path))
            shutil.rmtree(path)
        except IOError:
            # bug: http://code.activestate.com/lists/python-list/159050
            msg = "Unable to delete: {}".format(path)
            log.warning(msg)
    elif os.path.isfile(path):
        log.trace("Deleting '{}'...".format(path))
        os.remove(path)


def parse(text, path):
    """Parse a dictionary of data from formatted text.

    :param text: string containing dumped data
    :param path: file path to specify formatting

    :return: dictionary of data

    """
    ext = _get_ext(path)
    if ext in ['json']:
        data = _parse_json(text, path)
    elif ext in ['yml', 'yaml']:
        data = _parse_yaml(text, path)
    else:
        log.warning("Unrecognized file extension: %s", ext)
        data = _parse_yaml(text, path)

    if not isinstance(data, dict):
        msg = "Invalid file contents: {}".format(path)
        raise exceptions.FileContentError(msg)

    return data


def _parse_json(text, path):
    try:
        return json.loads(text) or {}
    except json.JSONDecodeError as exc:
        msg = "Invalid JSON contents: {}:\n{}".format(path, exc)
        raise exceptions.FileContentError(msg) from None


def _parse_yaml(text, path):
    try:
        return yaml.load(text) or {}
    except yaml.error.YAMLError as exc:
        msg = "Invalid YAML contents: {}:\n{}".format(path, exc)
        raise exceptions.FileContentError(msg) from None


def dump(data, path):
    """Format a dictionary into a serialization format.

    :param text: dictionary of data to format
    :param path: file path to specify formatting

    :return: string of formatted data

    """
    ext = _get_ext(path)

    if ext in ['json']:
        return json.dumps(data, indent=4, sort_keys=True)

    if ext not in ['yml', 'yaml']:
        log.warning("Unrecognized file extension: %s", ext)

    return yaml.dump(data, default_flow_style=False, allow_unicode=True)


def _get_ext(path):
    if '.' in path:
        return path.split('.')[-1].lower()
    else:
        return 'yml'



"""Custom YAML representers."""

from collections import OrderedDict

import yaml


class LiteralString(str):
    """Custom type for strings which should be dumped in the literal style."""


def represent_literalstring(dumper, data):
    return dumper.represent_scalar('tag:yaml.org,2002:str', data,
                                   style='|' if data else '')


def represent_ordereddict(dumper, data):
    value = []

    for item_key, item_value in data.items():
        node_key = dumper.represent_data(item_key)
        node_value = dumper.represent_data(item_value)

        value.append((node_key, node_value))

    return yaml.nodes.MappingNode('tag:yaml.org,2002:map', value)


yaml.add_representer(OrderedDict, represent_ordereddict)
yaml.add_representer(LiteralString, represent_literalstring)



# pylint: disable=missing-docstring,unused-variable,expression-not-assigned

import pytest
from expecter import expect

from yorm.decorators import attr
from yorm.types.standard import Integer, String, Float
from yorm.types.extended import (NullableString, Markdown,
                                 AttributeDictionary, SortedList)


def describe_nullable_string():

    def describe_to_value():

        def it_allows_none():
            expect(NullableString.to_value(None)).is_none()

    def describe_to_data():

        def it_allows_none():
            expect(NullableString.to_data(None)).is_none()


def describe_markdown():

    obj = "This is **the** sentence."
    data_value = [
        (obj, obj),
        (None, ""),
        (['a', 'b', 'c'], "a, b, c"),
        ("This is\na sentence.", "This is a sentence."),
        ("Sentence one.\nSentence two.", "Sentence one. Sentence two."),
    ]
    value_data = [
        (obj, obj + '\n'),
        ("Sentence one. Sentence two.", "Sentence one.\nSentence two.\n"),
        ("", ""),
        (" \t ", ""),
    ]

    def describe_to_value():

        @pytest.mark.parametrize("data,value", data_value)
        def it_converts_correctly(data, value):
            expect(Markdown.to_value(data)) == value

    def describe_to_data():

        @pytest.mark.parametrize("value,data", value_data)
        def it_converts_correctly(value, data):
            expect(Markdown.to_data(value)) == data


def describe_attribute_dictionary():

    @attr(var1=Integer)
    @attr(var2=String)
    class SampleAttributeDictionary(AttributeDictionary):
        """Sample attribute dictionary."""

    @pytest.fixture
    def converter():
        return SampleAttributeDictionary()

    def it_cannot_be_used_directly():
        with expect.raises(NotImplementedError):
            AttributeDictionary.to_value(None)
        with expect.raises(NotImplementedError):
            AttributeDictionary.to_data(None)

    def it_has_keys_available_as_attributes(converter):
        value = converter.to_value({'var1': 1, 'var2': "2"})
        expect(value.var1) == 1
        expect(value.var2) == "2"


def describe_sorted_list():

    @attr(all=Float)
    class SampleSortedList(SortedList):
        """Sample sorted list."""

    class UnknownSortedList(SortedList):
        """Sample list without a type."""

    @pytest.fixture
    def converter():
        return SampleSortedList()

    def it_cannot_be_used_directly():
        with expect.raises(NotImplementedError):
            SortedList.to_value(None)
        with expect.raises(NotImplementedError):
            SortedList.to_data(None)

    def it_cannot_be_subclassed_without_a_type():
        with expect.raises(NotImplementedError):
            UnknownSortedList.to_value(None)
        with expect.raises(NotImplementedError):
            UnknownSortedList.to_data(None)

    def describe_to_data():

        def it_sorts(converter):
            data = converter.to_data([4, 2, 0, 1, 3])
            expect(data) == [0.0, 1.0, 2.0, 3.0, 4.0]



"""Integration tests for file IO."""
# pylint: disable=missing-docstring,no-self-use,no-member,misplaced-comparison-constant


import logging

import pytest

import yorm
from yorm.types import Integer, String, Float, Boolean, Dictionary, List

from . import refresh_file_modification_times


# CLASSES ######################################################################


class EmptyDictionary(Dictionary):
    """Sample dictionary container."""


@yorm.attr(all=Integer)
class IntegerList(List):
    """Sample list container."""


@yorm.attr(array=IntegerList)
@yorm.attr(false=Boolean)
@yorm.attr(number_int=Integer)
@yorm.attr(number_real=Float)
@yorm.attr(object=EmptyDictionary)
@yorm.attr(string=String)
@yorm.attr(true=Boolean)
@yorm.sync("path/to/{self.category}/{self.name}.yml")
class SampleStandardDecorated:
    """Sample class using standard attribute types."""

    def __init__(self, name, category='default'):
        self.name = name
        self.category = category
        # https://docs.python.org/3.4/library/json.html#json.JSONDecoder
        self.object = {}
        self.array = []
        self.string = ""
        self.number_int = 0
        self.number_real = 0.0
        self.true = True
        self.false = False
        self.null = None

    def __repr__(self):
        return "<decorated {}>".format(id(self))


# TESTS ########################################################################


class TestCreate:
    """Integration tests for creating mapped classes."""

    def test_load_from_existing(self, tmpdir):
        """Verify attributes are updated from an existing file."""
        tmpdir.chdir()
        sample = SampleStandardDecorated('sample')
        sample2 = SampleStandardDecorated('sample')
        assert sample2.__mapper__.path == sample.__mapper__.path

        refresh_file_modification_times()

        logging.info("Changing values in object 1...")
        sample.array = [0, 1, 2]
        sample.string = "Hello, world!"
        sample.number_int = 42
        sample.number_real = 4.2
        sample.true = True
        sample.false = False

        logging.info("Reading changed values in object 2...")
        assert [0, 1, 2] == sample2.array
        assert "Hello, world!" == sample2.string
        assert 42 == sample2.number_int
        assert 4.2 == sample2.number_real
        assert True is sample2.true
        assert False is sample2.false


class TestDelete:
    """Integration tests for deleting files."""

    def test_read(self, tmpdir):
        """Verify a deleted file cannot be read from."""
        tmpdir.chdir()
        sample = SampleStandardDecorated('sample')
        sample.__mapper__.delete()

        with pytest.raises(FileNotFoundError):
            print(sample.string)

        with pytest.raises(FileNotFoundError):
            sample.string = "def456"

    def test_write(self, tmpdir):
        """Verify a deleted file cannot be written to."""
        tmpdir.chdir()
        sample = SampleStandardDecorated('sample')
        sample.__mapper__.delete()

        with pytest.raises(FileNotFoundError):
            sample.string = "def456"

    def test_multiple(self, tmpdir):
        """Verify a deleted file can be deleted again."""
        tmpdir.chdir()
        sample = SampleStandardDecorated('sample')
        sample.__mapper__.delete()
        sample.__mapper__.delete()


class TestUpdate:
    """Integration tests for updating files/object."""

    def test_automatic_save_after_first_modification(self, tmpdir):
        tmpdir.chdir()
        sample = SampleStandardDecorated('sample')
        assert "number_int: 0\n" in sample.__mapper__.text

        sample.number_int = 42
        assert "number_int: 42\n" in sample.__mapper__.text

        sample.__mapper__.text = "number_int: true\n"
        assert 1 is sample.number_int
        assert "number_int: 1\n" in sample.__mapper__.text

    def test_automatic_save_after_first_modification_on_list(self, tmpdir):
        tmpdir.chdir()
        sample = SampleStandardDecorated('sample')
        assert "array: []\n" in sample.__mapper__.text

        sample.array.append(42)
        assert "array:\n- 42\n" in sample.__mapper__.text

        sample.__mapper__.text = "array: [true]\n"
        iter(sample.array)
        assert "array:\n- 1\n" in sample.__mapper__.text



"""Integration tests for nested attributes."""
# pylint: disable=missing-docstring,no-self-use,attribute-defined-outside-init,no-member
# pylint: disable=unused-variable,misplaced-comparison-constant

from unittest.mock import patch
import logging

import pytest
from expecter import expect

import yorm

from . import strip


@yorm.attr(status=yorm.types.Boolean)
@yorm.attr(checked=yorm.types.Integer)
class StatusDictionary(yorm.types.Dictionary):

    def __init__(self, status, checked):
        self.status = status
        self.checked = checked
        if self.checked == 42:
            raise RuntimeError


@yorm.attr(all=yorm.types.Float)
class NestedList3(yorm.types.List):

    def __repr__(self):
        return "<nested-list-3 {}>".format((id(self)))


@yorm.attr(nested_list_3=NestedList3)
@yorm.attr(number=yorm.types.Float)
class NestedDictionary3(yorm.types.AttributeDictionary):

    def __init__(self):
        super().__init__()
        self.number = 0
        self.nested_list_3 = []

    def __repr__(self):
        print(self.number)  # trigger a potential recursion issue
        return "<nested-dictionary-3 {}>".format((id(self)))


@yorm.attr(all=yorm.types.Float)
class NestedList2(yorm.types.List):

    def __repr__(self):
        return "<nested-list-2 {}>".format((id(self)))


@yorm.attr(nested_dict_3=NestedDictionary3)
@yorm.attr(number=yorm.types.Float)
class NestedDictionary2(yorm.types.AttributeDictionary):

    def __init__(self):
        super().__init__()
        self.number = 0

    def __repr__(self):
        print(self.number)  # trigger a potential recursion issue
        return "<nested-dictionary-2 {}>".format((id(self)))


@yorm.attr(nested_list_2=NestedList2)
@yorm.attr(number=yorm.types.Float)
class NestedDictionary(yorm.types.AttributeDictionary):

    def __init__(self):
        super().__init__()
        self.number = 0
        self.nested_list_2 = []

    def __repr__(self):
        return "<nested-dictionary {}>".format((id(self)))


@yorm.attr(all=NestedDictionary2)
class NestedList(yorm.types.List):

    def __repr__(self):
        return "<nested-list {}>".format((id(self)))


@yorm.attr(nested_dict=NestedDictionary)
@yorm.attr(nested_list=NestedList)
@yorm.sync("sample.yml")
class Top:

    def __init__(self):
        self.nested_list = []
        self.nested_dict = {}

    def __repr__(self):
        return "<top {}>".format((id(self)))


@patch('yorm.settings.fake', True)
class TestNestedOnce:

    def test_append_triggers_save(self):
        top = Top()
        logging.info("Appending dictionary to list...")
        top.nested_list.append({'number': 1})
        assert strip("""
        nested_dict:
          nested_list_2: []
          number: 0.0
        nested_list:
        - nested_dict_3:
            nested_list_3: []
            number: 0.0
          number: 1.0
        """) == top.__mapper__.text

    def test_set_by_index_triggers_save(self):
        top = Top()
        top.nested_list = [{'number': 1.5}]
        assert strip("""
        nested_dict:
          nested_list_2: []
          number: 0.0
        nested_list:
        - nested_dict_3:
            nested_list_3: []
            number: 0.0
          number: 1.5
        """) == top.__mapper__.text
        top.nested_list[0] = {'number': 1.6}
        assert strip("""
        nested_dict:
          nested_list_2: []
          number: 0.0
        nested_list:
        - nested_dict_3:
            nested_list_3: []
            number: 0.0
          number: 1.6
        """) == top.__mapper__.text

    def test_get_by_index_triggers_load(self):
        top = Top()
        top.__mapper__.text = strip("""
        nested_list:
        - number: 1.7
        """)
        assert 1.7 == top.nested_list[0].number

    def test_delete_index_triggers_save(self):
        top = Top()
        top.nested_list = [{'number': 1.8}, {'number': 1.9}]
        assert strip("""
        nested_dict:
          nested_list_2: []
          number: 0.0
        nested_list:
        - nested_dict_3:
            nested_list_3: []
            number: 0.0
          number: 1.8
        - nested_dict_3:
            nested_list_3: []
            number: 0.0
          number: 1.9
        """) == top.__mapper__.text
        del top.nested_list[0]
        assert strip("""
        nested_dict:
          nested_list_2: []
          number: 0.0
        nested_list:
        - nested_dict_3:
            nested_list_3: []
            number: 0.0
          number: 1.9
        """) == top.__mapper__.text

    def test_set_dict_as_attribute_triggers_save(self):
        top = Top()
        top.nested_dict.number = 2
        assert strip("""
        nested_dict:
          nested_list_2: []
          number: 2.0
        nested_list: []
        """) == top.__mapper__.text


@patch('yorm.settings.fake', True)
class TestNestedTwice:

    def test_nested_list_item_value_change_triggers_save(self):
        top = Top()
        top.nested_list = [{'number': 3}]
        assert strip("""
        nested_dict:
          nested_list_2: []
          number: 0.0
        nested_list:
        - nested_dict_3:
            nested_list_3: []
            number: 0.0
          number: 3.0
        """) == top.__mapper__.text
        top.nested_list[0].number = 4
        assert strip("""
        nested_dict:
          nested_list_2: []
          number: 0.0
        nested_list:
        - nested_dict_3:
            nested_list_3: []
            number: 0.0
          number: 4.0
        """) == top.__mapper__.text

    def test_nested_dict_item_value_change_triggers_save(self):
        top = Top()
        top.nested_dict = {'nested_list_2': [5]}
        assert strip("""
        nested_dict:
          nested_list_2:
          - 5.0
          number: 0.0
        nested_list: []
        """) == top.__mapper__.text
        top.nested_dict.nested_list_2.append(6)
        assert strip("""
        nested_dict:
          nested_list_2:
          - 5.0
          - 6.0
          number: 0.0
        nested_list: []
        """) == top.__mapper__.text

    def test_dict_in_list_value_change_triggers_save(self):
        top = Top()
        top.nested_list.append(None)
        top.nested_list[0].nested_dict_3.number = 8
        assert strip("""
        nested_dict:
          nested_list_2: []
          number: 0.0
        nested_list:
        - nested_dict_3:
            nested_list_3: []
            number: 8.0
          number: 0.0
        """) == top.__mapper__.text

    def test_list_in_dict_append_triggers_save(self):
        top = Top()
        top.nested_list.append(None)
        top.nested_list.append(None)
        for nested_dict_2 in top.nested_list:
            nested_dict_2.number = 9
            nested_dict_2.nested_dict_3.nested_list_3.append(10)
        assert strip("""
        nested_dict:
          nested_list_2: []
          number: 0.0
        nested_list:
        - nested_dict_3:
            nested_list_3:
            - 10.0
            number: 0.0
          number: 9.0
        - nested_dict_3:
            nested_list_3:
            - 10.0
            number: 0.0
          number: 9.0
        """) == top.__mapper__.text


def describe_aliases():

    @pytest.fixture
    def sample(tmpdir):
        cls = type('Sample', (), {})
        path = str(tmpdir.join("sample.yml"))
        attrs = dict(var4=NestedList3, var5=StatusDictionary)
        return yorm.sync(cls(), path, attrs)

    def _log_ref(name, var, ref):
        logging.info("%s: %r", name, var)
        logging.info("%s_ref: %r", name, ref)
        logging.info("%s ID: %s", name, id(var))
        logging.info("%s_ref ID: %s", name, id(ref))
        assert id(ref) == id(var)
        assert ref == var

    def test_alias_list(sample):
        var4_ref = sample.var4
        _log_ref('var4', sample.var4, var4_ref)
        assert [] == sample.var4

        logging.info("Appending 42 to var4_ref...")
        var4_ref.append(42)
        _log_ref('var4', sample.var4, var4_ref)
        assert [42] == sample.var4

        logging.info("Appending 2015 to var4_ref...")
        var4_ref.append(2015)
        assert [42, 2015] == sample.var4

    def test_alias_dict(sample):
        var5_ref = sample.var5
        _log_ref('var5', sample.var5, var5_ref)
        assert {'status': False, 'checked': 0} == sample.var5

        logging.info("Setting status=True in var5_ref...")
        var5_ref['status'] = True
        _log_ref('var5', sample.var5, var5_ref)
        assert {'status': True, 'checked': 0} == sample.var5

        logging.info("Setting status=False in var5_ref...")
        var5_ref['status'] = False
        _log_ref('var5', sample.var5, var5_ref)
        assert {'status': False, 'checked': 0} == sample.var5

    def test_alias_dict_in_list():
        top = Top()
        top.nested_list.append(None)
        ref1 = top.nested_list[0]
        ref2 = top.nested_list[0].nested_dict_3
        ref3 = top.nested_list[0].nested_dict_3.nested_list_3
        assert id(ref1) == id(top.nested_list[0])
        assert id(ref2) == id(top.nested_list[0].nested_dict_3)
        assert id(ref3) == id(top.nested_list[0].nested_dict_3.nested_list_3)

    def test_alias_list_in_dict():
        top = Top()
        logging.info("Updating nested attribute...")
        top.nested_dict.number = 1
        logging.info("Grabbing refs...")
        ref1 = top.nested_dict
        ref2 = top.nested_dict.nested_list_2
        assert id(ref1) == id(top.nested_dict)
        assert id(ref2) == id(top.nested_dict.nested_list_2)

    def test_custom_init_is_invoked(sample):
        sample.__mapper__.text = "var5:\n  checked: 42"
        with expect.raises(RuntimeError):
            print(sample.var5)



# pylint: disable=redefined-outer-name,expression-not-assigned,attribute-defined-outside-init,no-member

from expecter import expect

import yorm
from yorm.types import List, Float

from . import strip


@yorm.attr(things=List.of_type(Float))
@yorm.sync("example.yml")
class Example:
    """An example class mapping a list using the shortened syntax."""


def test_list_mapping_using_shortened_syntax():
    obj = Example()
    obj.things.append(1)
    obj.things.append(2.0)
    obj.things.append("3")

    expect(obj.__mapper__.text) == strip("""
    things:
    - 1.0
    - 2.0
    - 3.0
    """)



"""Unit tests for the package."""

import os
import time
import logging

import expecter


def is_none(x):
    return x is None


def is_true(x):
    return x is True


def is_false(x):
    return x is False


def exists(x):
    return os.path.exists(x)


def missing(x):
    return not os.path.exists(x)


expecter.add_expectation(is_none)
expecter.add_expectation(is_true)
expecter.add_expectation(is_false)
expecter.add_expectation(exists)
expecter.add_expectation(missing)


def strip(text, tabs=None, end='\n'):
    """Strip leading whitespace indentation on multiline string literals."""
    lines = []

    for line in text.strip().splitlines():
        if not tabs:
            tabs = line.count(' ' * 4)
        lines.append(line.replace(' ' * tabs * 4, '', 1))

    return '\n'.join(lines) + end


def refresh_file_modification_times(seconds=1.1):
    """Sleep to allow file modification times to refresh."""
    logging.info("Delaying for %s second%s...", seconds,
                 "" if seconds == 1 else "s")
    time.sleep(seconds)



"""Integration tests for the package namespace."""
# pylint: disable=missing-docstring,no-self-use,unused-variable


def test_top():
    import yorm
    assert yorm.bases
    assert yorm.types.Integer
    assert yorm.types.extended.Markdown


def test_from_top_constants():
    from yorm import UUID


def test_from_top_clases():
    from yorm import Mappable
    from yorm import Converter, Container


def test_from_top_decorators():
    from yorm import sync
    from yorm import sync_instances
    from yorm import sync_object
    from yorm import attr


def test_from_top_utilities():
    from yorm import create
    from yorm import find
    from yorm import match
    from yorm import load
    from yorm import save
    from yorm import delete


def test_from_nested():
    from yorm.types import Integer
    from yorm.types.standard import String
    from yorm.types.extended import Markdown
    from yorm.types.containers import List



"""Convertible classes for builtin immutable types."""

import logging

from .. import exceptions
from ..bases import Converter

log = logging.getLogger(__name__)


class Object(Converter):
    """Base class for immutable types."""

    TYPE = None  # type for inferred types (set in subclasses)
    DEFAULT = None  # default value for conversion (set in subclasses)

    @classmethod
    def create_default(cls):
        return cls.DEFAULT

    @classmethod
    def to_value(cls, obj):
        return obj

    @classmethod
    def to_data(cls, obj):
        return cls.to_value(obj)


class String(Object):
    """Convertible for the `str` type."""

    TYPE = str
    DEFAULT = ""

    @classmethod
    def to_value(cls, obj):
        if isinstance(obj, cls.TYPE):
            return obj
        elif obj is True:
            return "true"
        elif obj is False:
            return "false"
        elif obj:
            try:
                return ', '.join(str(item) for item in obj)
            except TypeError:
                return str(obj)
        else:
            return cls.DEFAULT

    @classmethod
    def to_data(cls, obj):
        value = cls.to_value(obj)
        return cls._optimize_for_quoting(value)

    @staticmethod
    def _optimize_for_quoting(value):
        if value == "true":
            return True
        if value == "false":
            return False
        for number_type in (int, float):
            try:
                return number_type(value)
            except (TypeError, ValueError):
                continue
        return value


class Integer(Object):
    """Convertible for the `int` type."""

    TYPE = int
    DEFAULT = 0

    @classmethod
    def to_value(cls, obj):
        if all((isinstance(obj, cls.TYPE),
                obj is not True,
                obj is not False)):
            return obj
        elif obj:
            try:
                return int(obj)
            except ValueError:
                return int(float(obj))
        else:
            return cls.DEFAULT


class Float(Object):
    """Convertible for the `float` type."""

    TYPE = float
    DEFAULT = 0.0

    @classmethod
    def to_value(cls, obj):
        if isinstance(obj, cls.TYPE):
            return obj
        elif obj:
            return float(obj)
        else:
            return cls.DEFAULT


class Boolean(Object):
    """Convertible for the `bool` type."""

    TYPE = bool
    DEFAULT = False

    FALSY = ('false', 'f', 'no', 'n', 'disabled', 'off', '0')

    @classmethod
    def to_value(cls, obj):
        if isinstance(obj, str) and obj.lower().strip() in cls.FALSY:
            return False
        elif obj is not None:
            return bool(obj)
        else:
            return cls.DEFAULT


def match(name, data, nested=False):
    """Determine the appropriate converter for new data."""
    nested = " nested" if nested else ""
    msg = "Determining converter for new%s: '%s' = %r"
    log.debug(msg, nested, name, repr(data))

    types = Object.__subclasses__()  # pylint: disable=no-member
    log.trace("Converter options: {}".format(types))

    for converter in types:
        if converter.TYPE and type(data) == converter.TYPE:  # pylint: disable=unidiomatic-typecheck
            log.debug("Matched converter: %s", converter)
            log.info("New%s attribute: %s", nested, name)
            return converter

    if data is None or isinstance(data, (dict, list)):
        log.info("Default converter: %s", Object)
        log.warning("New%s attribute with unknown type: %s", nested, name)
        return Object

    msg = "No converter available for: {}".format(data)
    raise exceptions.FileContentError(msg)



"""Integration tests for the package."""
# pylint: disable=missing-docstring,no-self-use,no-member,misplaced-comparison-constant,attribute-defined-outside-init

import logging

import yorm
from yorm.types import Object, String, Integer, Float, Boolean
from yorm.types import Markdown, Dictionary, List

from . import strip, refresh_file_modification_times

log = logging.getLogger(__name__)


# CLASSES ######################################################################


class EmptyDictionary(Dictionary):
    """Sample dictionary container."""


@yorm.attr(all=Integer)
class IntegerList(List):
    """Sample list container."""


class SampleStandard:
    """Sample class using standard attribute types."""

    def __init__(self):
        # https://docs.python.org/3.4/library/json.html#json.JSONDecoder
        self.object = {}
        self.array = []
        self.string = ""
        self.number_int = 0
        self.number_real = 0.0
        self.truthy = True
        self.falsey = False
        self.null = None

    def __repr__(self):
        return "<standard {}>".format(id(self))


@yorm.attr(array=IntegerList)
@yorm.attr(falsey=Boolean)
@yorm.attr(number_int=Integer)
@yorm.attr(number_real=Float)
@yorm.attr(object=EmptyDictionary)
@yorm.attr(string=String)
@yorm.attr(truthy=Boolean)
@yorm.sync("path/to/{self.category}/{self.name}.yml")
class SampleStandardDecorated:
    """Sample class using standard attribute types."""

    def __init__(self, name, category='default'):
        self.name = name
        self.category = category
        # https://docs.python.org/3.4/library/json.html#json.JSONDecoder
        self.object = {}
        self.array = []
        self.string = ""
        self.number_int = 0
        self.number_real = 0.0
        self.truthy = True
        self.falsey = False
        self.null = None

    def __repr__(self):
        return "<decorated {}>".format(id(self))


@yorm.attr(label=String)
@yorm.attr(status=Boolean)
class StatusDictionary(Dictionary):
    """Sample dictionary container."""


@yorm.attr(all=StatusDictionary)
class StatusDictionaryList(List):
    """Sample list container."""


class Level(String):
    """Sample custom attribute."""

    @classmethod
    def to_data(cls, obj):
        value = cls.to_value(obj)
        count = value.split('.')
        if count == 0:
            return int(value)
        elif count == 1:
            return float(value)
        else:
            return value


@yorm.sync("path/to/directory/{UUID}.yml", attrs={'level': Level})
class SampleCustomDecorated:
    """Sample class using custom attribute types."""

    def __init__(self, name):
        self.name = name
        self.level = '1.0'

    def __repr__(self):
        return "<custom {}>".format(id(self))


@yorm.attr(string=String)
@yorm.sync("sample.yml", auto_save=False)
class SampleDecoratedAutoOff:
    """Sample class with automatic storage turned off."""

    def __init__(self):
        self.string = ""

    def __repr__(self):
        return "<auto save off {}>".format(id(self))


@yorm.sync("sample.yml", auto_track=True)
class SampleEmptyDecorated:
    """Sample class using standard attribute types."""

    def __repr__(self):
        return "<empty {}>".format(id(self))


class SampleExtended:
    """Sample class using extended attribute types."""

    def __init__(self):
        self.text = ""

    def __repr__(self):
        return "<extended {}>".format(id(self))


class SampleNested:
    """Sample class using nested attribute types."""

    def __init__(self):
        self.count = 0
        self.results = []

    def __repr__(self):
        return "<nested {}>".format(id(self))

# TESTS ########################################################################


class TestStandard:
    """Integration tests for standard attribute types."""

    @yorm.attr(status=yorm.types.Boolean)
    class StatusDictionary(Dictionary):
        pass

    def test_decorator(self, tmpdir):
        """Verify standard attribute types dump/parse correctly (decorator)."""
        tmpdir.chdir()
        sample = SampleStandardDecorated('sample')
        assert "path/to/default/sample.yml" == sample.__mapper__.path

        log.info("Checking object default values...")
        assert {} == sample.object
        assert [] == sample.array
        assert "" == sample.string
        assert 0 == sample.number_int
        assert 0.0 == sample.number_real
        assert True is sample.truthy
        assert False is sample.falsey
        assert None is sample.null

        log.info("Changing object values...")
        sample.object = {'key2': 'value'}
        sample.array = [0, 1, 2]
        sample.string = "Hello, world!"
        sample.number_int = 42
        sample.number_real = 4.2
        sample.truthy = False
        sample.falsey = True

        log.info("Checking file contents...")
        assert strip("""
        array:
        - 0
        - 1
        - 2
        falsey: true
        number_int: 42
        number_real: 4.2
        object: {}
        string: Hello, world!
        truthy: false
        """) == sample.__mapper__.text

        log.info("Changing file contents...")
        refresh_file_modification_times()
        sample.__mapper__.text = strip("""
        array: [4, 5, 6]
        falsey: null
        number_int: 42
        number_real: '4.2'
        object: {'status': false}
        string: "abc"
        truthy: null
        """)

        log.info("Checking object values...")
        assert {'status': False} == sample.object
        assert [4, 5, 6] == sample.array
        assert "abc" == sample.string
        assert 42 == sample.number_int
        assert 4.2 == sample.number_real
        assert False is sample.truthy
        assert False is sample.falsey

    def test_function(self, tmpdir):
        """Verify standard attribute types dump/parse correctly (function)."""
        tmpdir.chdir()
        _sample = SampleStandard()
        attrs = {'object': self.StatusDictionary,
                 'array': IntegerList,
                 'string': String,
                 'number_int': Integer,
                 'number_real': Float,
                 'truthy': Boolean,
                 'falsey': Boolean}
        sample = yorm.sync(_sample, "path/to/directory/sample.yml", attrs)
        assert "path/to/directory/sample.yml" == sample.__mapper__.path

        # check defaults
        assert {'status': False} == sample.object
        assert [] == sample.array
        assert "" == sample.string
        assert 0 == sample.number_int
        assert 0.0 == sample.number_real
        assert True is sample.truthy
        assert False is sample.falsey
        assert None is sample.null

        # change object values
        sample.object = {'key': 'value'}
        sample.array = [1, 2, 3]
        sample.string = "Hello, world!"
        sample.number_int = 42
        sample.number_real = 4.2
        sample.truthy = None
        sample.falsey = 1

        # check file values
        assert strip("""
        array:
        - 1
        - 2
        - 3
        falsey: true
        number_int: 42
        number_real: 4.2
        object:
          status: false
        string: Hello, world!
        truthy: false
        """) == sample.__mapper__.text

    def test_function_to_json(self, tmpdir):
        """Verify standard attribute types dump/parse correctly (function)."""
        tmpdir.chdir()
        _sample = SampleStandard()
        attrs = {'object': self.StatusDictionary,
                 'array': IntegerList,
                 'string': String,
                 'number_int': Integer,
                 'number_real': Float,
                 'truthy': Boolean,
                 'falsey': Boolean}
        sample = yorm.sync(_sample, "path/to/directory/sample.json", attrs)
        assert "path/to/directory/sample.json" == sample.__mapper__.path

        # check defaults
        assert {'status': False} == sample.object
        assert [] == sample.array
        assert "" == sample.string
        assert 0 == sample.number_int
        assert 0.0 == sample.number_real
        assert True is sample.truthy
        assert False is sample.falsey
        assert None is sample.null

        # change object values
        sample.object = {'key': 'value'}
        sample.array = [1, 2, 3]
        sample.string = "Hello, world!"
        sample.number_int = 42
        sample.number_real = 4.2
        sample.truthy = None
        sample.falsey = 1

        # check file values
        assert strip("""
        {
            "array": [
                1,
                2,
                3
            ],
            "falsey": true,
            "number_int": 42,
            "number_real": 4.2,
            "object": {
                "status": false
            },
            "string": "Hello, world!",
            "truthy": false
        }
        """, tabs=2, end='') == sample.__mapper__.text

    def test_auto_off(self, tmpdir):
        """Verify file updates are disabled with auto save off."""
        tmpdir.chdir()
        sample = SampleDecoratedAutoOff()

        sample.string = "hello"
        assert "" == sample.__mapper__.text

        sample.__mapper__.auto_save = True
        sample.string = "world"

        assert strip("""
        string: world
        """) == sample.__mapper__.text


class TestContainers:
    """Integration tests for attribute containers."""

    def test_nesting(self, tmpdir):
        """Verify standard attribute types can be nested."""
        tmpdir.chdir()
        _sample = SampleNested()
        attrs = {'count': Integer,
                 'results': StatusDictionaryList}
        sample = yorm.sync(_sample, "sample.yml", attrs, auto_track=True)

        # check defaults
        assert 0 == sample.count
        assert [] == sample.results

        # change object values
        sample.count = 5
        sample.results = [{'status': False, 'label': "abc"},
                          {'status': None, 'label': None},
                          {'label': "def"},
                          {'status': True},
                          {}]

        # check file values
        assert strip("""
        count: 5
        results:
        - label: abc
          status: false
        - label: ''
          status: false
        - label: def
          status: false
        - label: ''
          status: true
        - label: ''
          status: false
        """) == sample.__mapper__.text

        # change file values
        refresh_file_modification_times()
        sample.__mapper__.text = strip("""
        count: 3
        other: 4.2
        results:
        - label: abc
        - label: null
          status: false
        - status: true
        """)

        # check object values
        assert 3 == sample.count
        assert 4.2 == sample.other
        assert [{'label': 'abc', 'status': False},
                {'label': '', 'status': False},
                {'label': '', 'status': True}] == sample.results

    def test_objects(self, tmpdir):
        """Verify containers are treated as objects when added."""
        tmpdir.chdir()
        sample = SampleEmptyDecorated()

        # change file values
        refresh_file_modification_times()
        sample.__mapper__.text = strip("""
        object: {'key': 'value'}
        array: [1, '2', '3.0']
        """)

        # (a mapped attribute must be read first to trigger retrieving)
        sample.__mapper__.load()

        # check object values
        assert {'key': 'value'} == sample.object
        assert [1, '2', '3.0'] == sample.array

        # check object types
        assert Object == sample.__mapper__.attrs['object']
        assert Object == sample.__mapper__.attrs['array']


class TestExtended:
    """Integration tests for extended attribute types."""

    def test_function(self, tmpdir):
        """Verify extended attribute types dump/parse correctly."""
        tmpdir.chdir()
        _sample = SampleExtended()
        attrs = {'text': Markdown}
        sample = yorm.sync(_sample, "path/to/directory/sample.yml", attrs)

        # check defaults
        assert "" == sample.text

        # change object values
        refresh_file_modification_times()
        sample.text = strip("""
        This is the first sentence. This is the second sentence.
        This is the third sentence.
        """)

        # check file values
        assert strip("""
        text: |
          This is the first sentence.
          This is the second sentence.
          This is the third sentence.
        """) == sample.__mapper__.text

        # change file values
        refresh_file_modification_times()
        sample.__mapper__.text = strip("""
        text: |
          This is a
          sentence.
        """)

        # check object values
        assert "This is a sentence." == sample.text


class TestCustom:
    """Integration tests for custom attribute types."""

    def test_decorator(self, tmpdir):
        """Verify custom attribute types dump/parse correctly."""
        tmpdir.chdir()
        sample = SampleCustomDecorated('sample')

        # check defaults
        assert '1.0' == sample.level

        # change values
        sample.level = '1.2.3'

        # check file values
        assert strip("""
        level: 1.2.3
        """) == sample.__mapper__.text

        # change file values
        refresh_file_modification_times()
        sample.__mapper__.text = strip("""
        level: 1
        """)

        # check object values
        assert '1' == sample.level



# pylint: disable=missing-docstring,unused-variable,expression-not-assigned

import pytest
from expecter import expect

from yorm.types import Object, String, Integer, Float, Boolean


def describe_object():

    obj = None
    pairs = [
        (obj, obj),
        (None, None),
        (1, 1),
        (4.2, 4.2),
        (['a', 'b', 'c'], ['a', 'b', 'c']),
    ]

    def describe_to_value():

        @pytest.mark.parametrize("first,second", pairs)
        def it_converts_correctly(first, second):
            expect(Object.to_value(first)) == second

    def describe_to_data():

        @pytest.mark.parametrize("first,second", pairs)
        def it_converts_correctly(first, second):
            expect(Object.to_data(first)) == second


def describe_string():

    obj = "Hello, world!"
    pairs = [
        (obj, obj),
        (None, ""),
        ("1.2.3", "1.2.3"),
        (['a', 'b', 'c'], "a, b, c"),
    ]
    pairs_to_value = pairs + [
        (1, "1"),
        (4.2, "4.2"),
        (False, "false"),
        (True, "true"),
    ]
    pairs_to_data = pairs + [
        (42, 42),
        (4.2, 4.2),
        ("true", True),
        ("false", False),
    ]

    def describe_to_value():

        @pytest.mark.parametrize("first,second", pairs_to_value)
        def it_converts_correctly(first, second):
            expect(String.to_value(first)) == second

    def describe_to_data():

        @pytest.mark.parametrize("first,second", pairs_to_data)
        def it_converts_correctly(first, second):
            expect(String.to_data(first)) == second


def describe_integer():

    obj = 42
    pairs = [
        (obj, obj),
        (None, 0),
        ("1", 1),
        ("1.1", 1),
        (True, 1),
        (False, 0),
    ]

    def describe_to_value():

        @pytest.mark.parametrize("first,second", pairs)
        def it_converts_correctly(first, second):
            expect(Integer.to_value(first)) == second

        def it_raises_an_exception_when_unable_to_convert():
            with expect.raises(ValueError):
                Integer.to_value("abc")

    def describe_to_data():

        @pytest.mark.parametrize("first,second", pairs)
        def it_converts_correctly(first, second):
            expect(Integer.to_data(first)) == second


def describe_float():

    obj = 4.2
    pairs = [
        (obj, obj),
        (None, 0.0),
        ("1.0", 1.0),
        ("1.1", 1.1),
    ]

    def describe_to_value():

        @pytest.mark.parametrize("first,second", pairs)
        def it_converts_correctly(first, second):
            expect(Float.to_value(first)) == second

        def it_raises_an_exception_when_unable_to_convert():
            with expect.raises(ValueError):
                Float.to_value("abc")

    def describe_to_data():

        @pytest.mark.parametrize("first,second", pairs)
        def it_converts_correctly(first, second):
            expect(Float.to_data(first)) == second


def describe_boolean():

    obj = True
    pairs = [
        (obj, obj),
        (None, False),
        (0, False),
        (1, True),
        ("", False),
        ("True", True),
        ("False", False),
        ("true", True),
        ("false", False),
        ("T", True),
        ("F", False),
        ("yes", True),
        ("no", False),
        ("Y", True),
        ("N", False),
        ("enabled", True),
        ("disabled", False),
        ("on", True),
        ("off", False),
        ("Hello, world!", True)
    ]

    def describe_to_value():

        @pytest.mark.parametrize("first,second", pairs)
        def it_converts_correctly(first, second):
            expect(Boolean.to_value(first)) == second

    def describe_to_data():

        @pytest.mark.parametrize("first,second", pairs)
        def it_converts_correctly(first, second):
            expect(Boolean.to_data(first)) == second



"""Converter classes."""

from abc import ABCMeta, abstractclassmethod, abstractmethod
import logging

from .. import common
from . import Mappable

log = logging.getLogger(__name__)


class Converter(metaclass=ABCMeta):
    """Base class for attribute converters."""

    @abstractclassmethod
    def create_default(cls):
        """Create a default value for an attribute."""
        raise NotImplementedError(common.OVERRIDE_MESSAGE)

    @abstractclassmethod
    def to_value(cls, data):
        """Convert parsed data to an attribute's value."""
        raise NotImplementedError(common.OVERRIDE_MESSAGE)

    @abstractclassmethod
    def to_data(cls, value):
        """Convert an attribute to data optimized for dumping."""
        raise NotImplementedError(common.OVERRIDE_MESSAGE)


class Container(Mappable, Converter, metaclass=ABCMeta):
    """Base class for mutable attribute converters."""

    @classmethod
    def create_default(cls):
        return cls.__new__(cls)

    @classmethod
    def to_value(cls, data):
        value = cls.create_default()
        value.update_value(data, auto_track=True)
        return value

    @abstractmethod
    def update_value(self, data, *, auto_track):  # pragma: no cover (abstract method)
        """Update the attribute's value from parsed data."""
        raise NotImplementedError(common.OVERRIDE_MESSAGE)

    def format_data(self):
        """Format the attribute to data optimized for dumping."""
        return self.to_data(self)



"""Converters for attributes."""

from .standard import Object, match
from .standard import Integer, Boolean, String, Float

from .containers import Dictionary, List

from .extended import NullableInteger, NullableBoolean
from .extended import NullableString, NullableFloat
from .extended import Markdown
from .extended import AttributeDictionary, SortedList



# pylint: disable=missing-docstring,no-self-use,misplaced-comparison-constant,abstract-class-instantiated

import pytest

from yorm.bases import Container


class TestContainer:
    """Unit tests for the `Container` class."""

    class MyContainer(Container):

        def __init__(self, number):
            from unittest.mock import MagicMock
            self.__mapper__ = MagicMock()
            self.value = number

        @classmethod
        def create_default(cls):
            return 1

        @classmethod
        def to_data(cls, value):
            return str(value.value)

        def update_value(self, data, *, auto_track=None):  # pylint: disable=unused-variable
            self.value += int(data)

    def test_container_class_cannot_be_instantiated(self):
        with pytest.raises(TypeError):
            Container()

    def test_container_instance_methods_can_be_called(self):
        container = self.MyContainer(42)
        assert 42 == container.value
        container.update_value(10)
        assert 52 == container.value
        assert "52" == container.format_data()



#!/usr/bin/env python

"""Setup script for the package."""

import os
import sys

import setuptools


PACKAGE_NAME = 'yorm'
MINIMUM_PYTHON_VERSION = 3, 3


def check_python_version():
    """Exit when the Python version is too low."""
    if sys.version_info < MINIMUM_PYTHON_VERSION:
        sys.exit("Python {}.{}+ is required.".format(*MINIMUM_PYTHON_VERSION))


def read_package_variable(key):
    """Read the value of a variable from the package without importing."""
    module_path = os.path.join(PACKAGE_NAME, '__init__.py')
    with open(module_path) as module:
        for line in module:
            parts = line.strip().split(' ')
            if parts and parts[0] == key:
                return parts[-1].strip("'")
    assert 0, "'{0}' not found in '{1}'".format(key, module_path)


def read_descriptions():
    """Build a description for the project from documentation files."""
    try:
        readme = open("README.rst").read()
        changelog = open("CHANGELOG.rst").read()
    except IOError:
        return "<placeholder>"
    else:
        return readme + '\n' + changelog


check_python_version()
setuptools.setup(
    name=read_package_variable('__project__'),
    version=read_package_variable('__version__'),

    description="Automatic object-YAML mapping for Python.",
    url='https://github.com/jacebrowning/yorm',
    author='Jace Browning',
    author_email='jacebrowning@gmail.com',

    packages=setuptools.find_packages(),

    entry_points={'console_scripts': []},

    long_description=read_descriptions(),
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Database',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Version Control',
        'Topic :: System :: Filesystems',
        'Topic :: Text Editors :: Text Processing',
    ],

    install_requires=open("requirements.txt").readlines(),
)



"""Converter classes for builtin container types."""

import logging

from .. import common
from ..bases import Container
from . import standard

log = logging.getLogger(__name__)


class Dictionary(Container, dict):
    """Base class for a dictionary of attribute types."""

    def __new__(cls, *args, **kwargs):
        if cls is Dictionary:
            msg = "Dictionary class must be subclassed to use"
            raise NotImplementedError(msg)
        return super().__new__(cls, *args, **kwargs)

    @classmethod
    def to_data(cls, value):
        value2 = cls.create_default()
        value2.update_value(value, auto_track=False)

        data = common.attrs[cls].__class__()
        for name, converter in common.attrs[cls].items():
            data[name] = converter.to_data(value2.get(name, None))

        return data

    def update_value(self, data, *, auto_track=True):
        cls = self.__class__
        value = cls.create_default()

        # Convert object attributes to a dictionary
        attrs = common.attrs[cls].copy()
        if isinstance(data, cls):
            dictionary = {}
            for k, v in data.items():
                if k in attrs:
                    dictionary[k] = v
            for k, v in data.__dict__.items():
                if k in attrs:
                    dictionary[k] = v
        else:
            dictionary = to_dict(data)

        # Map object attributes to types
        for name, data2 in dictionary.items():

            try:
                converter = attrs.pop(name)
            except KeyError:
                if auto_track:
                    converter = standard.match(name, data2, nested=True)
                    common.attrs[cls][name] = converter
                else:
                    msg = "Ignored unknown nested file attribute: %s = %r"
                    log.warning(msg, name, data2)
                    continue

            try:
                attr = self[name]
            except KeyError:
                attr = converter.create_default()

            if all((isinstance(attr, converter),
                    issubclass(converter, Container))):
                attr.update_value(data2, auto_track=auto_track)
            else:
                attr = converter.to_value(data2)

            value[name] = attr

        # Create default values for unmapped types
        for name, converter in attrs.items():
            value[name] = converter.create_default()
            msg = "Default value for missing nested object attribute: %s = %r"
            log.info(msg, name, value[name])

        # Execute custom __init__ validations
        try:
            cls(**value)
        except TypeError as exception:
            log.warning("%s: %s", cls.__name__, exception)

        # Apply the new value
        self.clear()
        self.update(value)


class List(Container, list):
    """Base class for a homogeneous list of attribute types."""

    def __new__(cls, *args, **kwargs):
        if cls is List:
            raise NotImplementedError("List class must be subclassed to use")
        if not cls.item_type:
            raise NotImplementedError("List subclass must specify item type")
        return super().__new__(cls, *args, **kwargs)

    @classmethod
    def of_type(cls, sub_class):
        name = sub_class.__name__ + cls.__name__
        new_class = type(name, (cls,), {})
        common.attrs[new_class][common.ALL] = sub_class
        return new_class

    @common.classproperty
    def item_type(cls):  # pylint: disable=no-self-argument
        """Get the converter class for all items."""
        return common.attrs[cls].get(common.ALL)

    @classmethod
    def to_data(cls, value):
        value2 = cls.create_default()
        value2.update_value(value, auto_track=False)

        data = []

        if value2:
            for item in value2:
                data.append(cls.item_type.to_data(item))  # pylint: disable=no-member

        return data

    def update_value(self, data, *, auto_track=True):
        cls = self.__class__
        value = cls.create_default()

        # Get the converter for all items
        converter = cls.item_type

        # Convert the parsed data
        for item in to_list(data):

            try:
                attr = self[len(value)]
            except IndexError:
                attr = converter.create_default()  # pylint: disable=no-member

            if all((isinstance(attr, converter),
                    issubclass(converter, Container))):
                attr.update_value(item, auto_track=auto_track)
            else:
                attr = converter.to_value(item)  # pylint: disable=no-member

            value.append(attr)

        # Apply the new value
        self[:] = value[:]


def to_dict(obj):
    """Convert a dictionary-like object to a dictionary.

    >>> to_dict({'key': 42})
    {'key': 42}

    >>> to_dict("key=42")
    {'key': '42'}

    >>> to_dict("key")
    {'key': None}

    >>> to_dict(None)
    {}

    """
    if isinstance(obj, dict):
        return obj
    elif isinstance(obj, str):
        text = obj.strip()
        parts = text.split('=')
        if len(parts) == 2:
            return {parts[0]: parts[1]}
        else:
            return {text: None}
    else:
        try:
            return obj.__dict__
        except AttributeError:
            return {}


def to_list(obj):
    """Convert a list-like object to a list.

    >>> to_list([1, 2, 3])
    [1, 2, 3]

    >>> to_list("a,b,c")
    ['a', 'b', 'c']

    >>> to_list("item")
    ['item']

    >>> to_list(None)
    []

    """
    if isinstance(obj, list):
        return obj
    elif isinstance(obj, str):
        text = obj.strip()
        if ',' in text and ' ' not in text:
            return text.split(',')
        else:
            return text.split()
    elif obj is not None:
        return [obj]
    else:
        return []



"""Functions to interact with mapped classes and instances."""

import inspect
import logging
import warnings

from . import common, exceptions

log = logging.getLogger(__name__)


def create(class_or_instance, *args, overwrite=False):
    """Create a new mapped object."""
    instance = _instantiate(class_or_instance, *args)
    mapper = common.get_mapper(instance, expected=True)

    if mapper.auto_create:
        msg = "'create' is called automatically with 'auto_create' enabled"
        warnings.warn(msg)

    if mapper.exists and not overwrite:
        msg = "{!r} already exists".format(mapper.path)
        raise exceptions.DuplicateMappingError(msg)

    return save(instance)


def find(class_or_instance, *args, create=False):  # pylint: disable=redefined-outer-name
    """Find a matching mapped object or return None."""
    instance = _instantiate(class_or_instance, *args)
    mapper = common.get_mapper(instance, expected=True)

    if mapper.exists:
        return instance
    elif create:
        return save(instance)
    else:
        return None


def match(cls, **kwargs):
    """Yield all matching mapped objects."""
    log.debug((cls, kwargs))
    raise NotImplementedError


def load(instance):
    """Force the loading of a mapped object's file."""
    mapper = common.get_mapper(instance, expected=True)

    warnings.warn("'load' is called automatically")

    mapper.load()

    return instance


def save(instance):
    """Save a mapped object to file."""
    mapper = common.get_mapper(instance, expected=True)

    if mapper.auto_save:
        msg = "'save' is called automatically with 'auto_save' enabled"
        warnings.warn(msg)

    if mapper.deleted:
        msg = "{!r} was deleted".format(mapper.path)
        raise exceptions.DeletedFileError(msg)

    if not mapper.exists:
        mapper.create()

    mapper.save()

    return instance


def delete(instance):
    """Delete a mapped object's file."""
    mapper = common.get_mapper(instance, expected=True)

    mapper.delete()

    return None


def _instantiate(class_or_instance, *args):
    if inspect.isclass(class_or_instance):
        instance = class_or_instance(*args)
    else:
        assert not args
        instance = class_or_instance

    return instance
